@@ -24,7 +24,7 @@
 #' @param export character vector of object names needed to evaluate \code{FUN}.
 #' @param packages character vector of the packages needed to run \code{FUN}.
 #' @param initialize should the process initialize a parameter-Score pair set? If \code{FALSE}, \code{leftOff} must be provided.
-#' @param initGrid user specified points to sample the target function, should
+#' @param initGrid user specified points to sample the scoring function, should
 #'   be a \code{data.frame} or \code{data.table} with identical column names as bounds.
 #' @param initPoints number of randomly chosen points to sample the
 #'   scoring function before Bayesian Optimization fitting the Gaussian Process.
@@ -48,9 +48,10 @@
 #' }
 #' @param stopImpatient a list containing \code{rounds} and \code{newAcq}, if \code{acq = "eips"} you
 #'   can switch the acquisition function to \code{newAcq} after \code{rounds} parameter-score pairs are found.
-#' @param kappa tunable parameter kappa of GP Upper Confidence Bound, to balance exploitation against exploration,
-#'   increasing kappa will incentivise exploration.
-#' @param eps tunable parameter epsilon of ei, eips and poi. Balances exploitation against exploration.
+#' @param kappa tunable parameter kappa of the upper confidence bound. Tunes exploitation/exploration.
+#'   Increasing kappa will increase the importance that variance (unexplored space) has, therefore
+#'   incentivising exploration.
+#' @param eps tunable parameter epsilon of ei, eips and poi. Tunes exploitation/exploration.
 #'   Increasing eps will make the "improvement" threshold higher.
 #' @param gsPoints integer that specifies how many initial points to try when searching for the optimal parameter set.
 #'   Increase this for a higher chance to find global optimum, at the expense of more time.
@@ -67,7 +68,8 @@
 #' @param verbose Whether or not to print progress. If 0, nothing will be printed.
 #'   If 1, progress will be printed. If 2, progress and information about new parameter-score pairs will be printed.
 #' @return A list containing details about the process:
-#' \item{GPlist}{The list of the gaussian process objects that were fit.}
+#' \item{GPs}{The last Gaussian Process run on the parameter-score pairs}
+#' \item{GPe}{If \code{acq = "eips"}, this contains the last Gaussian Process run on the parameter-elapsed time pairs}
 #' \item{acqMaximums}{The optimal parameters according to each gaussian process}
 #' \item{ScoreDT}{A list of all parameter-score pairs, as well as extra columns from FUN}
 #' \item{BestPars}{The best parameter set at each iteration}
@@ -150,7 +152,7 @@
 #'   , parallel = FALSE
 #'   , gsPoints = 50)
 #' }
-#' @importFrom data.table data.table setDT setcolorder := as.data.table copy .I
+#' @importFrom data.table data.table setDT setcolorder := as.data.table copy .I setnames is.data.table
 #' @importFrom utils head
 #' @importFrom GauPro GauPro_kernel_model Matern52 Matern32 Exponential Gaussian
 #' @export
@@ -183,27 +185,29 @@ BayesianOptimization <- function(
   StartT <- Sys.time()
 
   # Set counters and other helper objects
+  mco <- list(preschedule=FALSE)
   ParamNames <- names(bounds)
   packages <- unique(c('data.table',packages))
   setDT(initGrid)
   setDT(leftOff)
   Overwrites <- 0
   Iter <- 0
-  kern <- assignKern(kern,beta)
+  kerns <- assignKern(kern,beta)
   boundsDT <- data.table( N = ParamNames
                         , L = sapply(bounds, function(x) x[1])
                         , U = sapply(bounds, function(x) x[2])
+                        , R = sapply(bounds, function(x) x[2]) - sapply(bounds, function(x) x[1])
                         , C = sapply(bounds, function(x) class(x))
   )
 
-
-
+  
   # Define processing function
   ParMethod <- function(x) if(x) {`%dopar%`} else {`%do%`}
   `%op%` <- ParMethod(parallel)
   Workers <- getDoParWorkers()
 
-  # Ensure environment fidelity
+  
+  # Better to quit gracefully than not
   if (sum(acq == c("ucb","ei","eips","poi")) == 0) stop("Acquisition function not recognized")
   if (sum(stopImpatient$newAcq == c("ucb","ei","eips","poi")) == 0) stop("New acquisition function not recognized")
   if (!initialize & nrow(leftOff) == 0) stop("initialize cannot be FALSE if leftOff is not provided. Set initialize to TRUE and provide either initGrid or initPoints. You can provide leftOff AND initialize if you want.\n")
@@ -213,28 +217,30 @@ BayesianOptimization <- function(
   if (parallel & (Workers == 1)) stop("parallel is set to TRUE but no back end is registered.\n")
   if (!parallel & Workers > 1 & verbose > 0) cat("parallel back end is registered, but parallel is set to false. Process will not be run in parallel.\n")
   if (nrow(initGrid)>0) {
-    if (sum(sapply(ParamNames, CheckBounds,initGrid, bounds))>0) stop("initGrid not within bounds.")
+    if (sum(sapply(ParamNames, checkBounds,initGrid, bounds))>0) stop("initGrid not within bounds.")
   }
   if (nrow(leftOff) > 0){
-    if (sum(sapply(ParamNames, CheckBounds,leftOff, bounds))>0) stop("leftOff not within bounds.")
+    if (sum(sapply(ParamNames, checkBounds,leftOff, bounds))>0) stop("leftOff not within bounds.")
   }
   if (nrow(leftOff)+initialize*(initPoints+nrow(initGrid)) >= nIters) stop("Rows in initial set will be larger than nIters")
   if (verbose > 0 & bulkNew < Workers & parallel) cat("bulkNew is less than the threads registered on the parallel back end - process may not utilize all workers.\n")
 
 
-  # If an initGrid was specified, make that the initial process fit. If not, create one with initPoints combinations.
+  # If an initGrid was specified, make that the initial process fit. 
+  # If not, create one with initPoints combinations.
   # If leftOff was specified, append to the initialized table (if applicable)
   if (initialize){
 
       if (nrow(initGrid)>0){
         InitFeedParams <- initGrid
       } else{
-        InitFeedParams <- data.table(sapply(ParamNames,RandParams,initPoints, boundsDT))
+        InitFeedParams <- randParams(boundsDT, initPoints)
       }
 
       if (verbose > 0) cat("\nRunning initial scoring function",nrow(InitFeedParams),"times in",Workers,"thread(s).\n")
       sink("NUL")
       ScoreDT <- foreach( iter = 1:nrow(InitFeedParams)
+                        , .options.multicore = mco
                         , .combine = rbind
                         , .multicombine = TRUE
                         , .inorder = FALSE
@@ -246,10 +252,16 @@ BayesianOptimization <- function(
 
           Params <- InitFeedParams[get("iter"),]
           Elapsed <- system.time(Result <- do.call(what = FUN, args = as.list(Params)))
+          if(sum(names(Result) == "Score") == 0) stop("FUN must return list with element 'Score' at a minimum.")
           data.table(Params,Elapsed = Elapsed[[3]],as.data.table(Result))
 
       }
       sink()
+      
+      if (!is.data.table(ScoreDT)) {
+        cat("\nFUN failed to run with error list:\n"); print(ScoreDT)
+        stop("Stopping process.")
+      }
 
       # Append leftOff if its names match ScoreDT
       if (nrow(leftOff) > 0) {
@@ -282,14 +294,14 @@ BayesianOptimization <- function(
   }
 
   # Setup for iterations
-  GPlist <- list()
   acqMaximums <- data.table()
   scaleList <- list(score = max(abs(ScoreDT$Score)), elapsed = max(ScoreDT$Elapsed))
-  GP <- NULL
+  GPs <- NULL
+  GPe <- NULL
   BestPars <- data.table( "Iteration" = Iter
-                       , ScoreDT[which.max(get("Score")),c(ParamNames,"Score",extraRet),with = F]
-                       , elapsedSecs = round(difftime(Time,StartT,units = "secs"),0)
-                       )
+                        , ScoreDT[which.max(get("Score")),c(ParamNames,"Score",extraRet),with = F]
+                        , elapsedSecs = round(difftime(Time,StartT,units = "secs"),0)
+                        )
 
   # Start the iterative GP udpates
   while(nrow(ScoreDT) < nIters){
@@ -307,28 +319,55 @@ BayesianOptimization <- function(
         acq <- stopImpatient$newAcq
       }
 
+      
     # Fit GP
     newD <- ScoreDT[get("Iteration") == Iter-1,]
     if (verbose > 0) cat("\n  1) Fitting Gaussian process...")
-    GP <- updateGP( GP = GP
-                  , kern = kern
-                  , X = matrix(sapply(ParamNames, MinMaxScale, newD, boundsDT), nrow = nrow(newD))
-                  , Z = matrix(as.matrix(newD[,.(get("Score")/scaleList$score,Elapsed/scaleList$elapsed)]), nrow = nrow(newD))
-                  , acq = acq
-                  , scaleList = scaleList
-                  , parallel = parallel)
+    
+    # Fitting/Updating GauPro S6 class inside seperate function scope causes memory pointer problems.
+    X = matrix(as.matrix(minMaxScale(newD, boundsDT)), nrow = nrow(newD))
+    Z = matrix(as.matrix(newD[,.(get("Score")/scaleList$score,Elapsed/scaleList$elapsed)]), nrow = nrow(newD))
+    
+    if (is.null(GPs)) {
+      
+      GPs <- GauPro_kernel_model$new( X
+                                    , matrix(Z[,1])
+                                    , kernel = kerns
+                                    , parallel = FALSE
+                                    , useC = FALSE)
+      
+      if (acq == "eips") {
+        
+        kerne <- assignKern(kern,beta)
 
-    # Store GP in list
-    GPlist[[Iter]] <- GP
+        GPe <- GauPro_kernel_model$new( X
+                                      , matrix(Z[,2])
+                                      , kernel = kerne
+                                      , parallel = FALSE
+                                      , useC = FALSE)
+        
+      }
+    } else{
+      
+      # Update Score GP
+      GPs <- GPs$update(Xnew = X, Znew = as.matrix(Z[,1]), nug.update = TRUE)
+      
+      if (acq == "eips") {
+
+        GPe <- GPe$update(Xnew = X, Znew = as.matrix(Z[,2]), nug.update = TRUE)
+
+      } else GPe <- NULL
+    }
 
     # Create random points to initialize local maximum search.
-    LocalTries <- data.table(sapply(ParamNames,RandParams,gsPoints,boundsDT))
-    LocalTryMM <- data.table(sapply(ParamNames,MinMaxScale,LocalTries,boundsDT))
+    localTries <- randParams(boundsDT, gsPoints, FAIL = FALSE)
+    localTryMM <- minMaxScale(localTries, boundsDT)
 
     # Try gsPoints starting points to find parameter set that optimizes Acq
     if (verbose > 0) cat("\n  2) Running local optimum search...")
-    LocalOptims <- maxAcq( GP = GP
-                         , TryOver = LocalTryMM
+    LocalOptims <- maxAcq( GPs = GPs
+                         , GPe = GPe
+                         , TryOver = localTryMM
                          , acq = acq
                          , y_max = max(ScoreDT$Score/scaleList$score)
                          , kappa = kappa
@@ -337,22 +376,16 @@ BayesianOptimization <- function(
                          , parallel = parallel
                          , convThresh = convThresh
                          )
-
-    if (sum(LocalOptims$gradCount > 2) == 0) cat("\n  2a) WARNING - No initial points converged.\n      Process may just be sampling random points.\n      Try decreasing convThresh.")
+    
+    if (sum(LocalOptims$gradCount > 2) == 0) cat("\n  2a) optim function only took <3 steps.\n      Process may be sampling random points.\n      Try decreasing convThresh.")
 
     fromCluster <- applyCluster()
     acqMaximums <- rbind(acqMaximums, data.table("Iteration" = Iter, fromCluster$clusterPoints))
-    newScorePars <- sapply(ParamNames, UnMMScale, fromCluster$newSet, boundsDT)
 
-    if (runNew < 2) {
-      newScorePars <- as.data.table(as.list(newScorePars))
-    } else{
-      newScorePars <- data.table(newScorePars)
-    }
-
-    if (verbose > 0) cat("\n  3) Running scoring function",nrow(newScorePars),"times in",Workers,"thread(s)...\n")
+    if (verbose > 0) cat("\n  3) Running scoring function",nrow(fromCluster$newSet),"times in",Workers,"thread(s)...\n")
     sink("NUL")
-    NewResults <- foreach( iter = 1:nrow(newScorePars)
+    NewResults <- foreach( iter = 1:nrow(fromCluster$newSet)
+                         , .options.multicore = mco
                          , .combine = rbind
                          , .multicombine = TRUE
                          , .inorder = FALSE
@@ -362,13 +395,18 @@ BayesianOptimization <- function(
                          , .export = export
                          ) %op% {
 
-            Params <- newScorePars[get("iter"),]
+            Params <- fromCluster$newSet[get("iter"),]
             Elapsed <- system.time(Result <- do.call(what = FUN, args = as.list(Params)))
-            data.table(newScorePars[get("iter"),], Elapsed = Elapsed[[3]], as.data.table(Result))
-
-            }
+            data.table(fromCluster$newSet[get("iter"),], Elapsed = Elapsed[[3]], as.data.table(Result))
+            
+    }
     sink()
-
+    
+    if (!is.data.table(NewResults)) {
+      cat("\nFUN failed to run with error list:\n"); print(NewResults)
+      stop("Stopping process.")
+    }
+    
     Time <- Sys.time()
 
     # Print updates on parameter-score search
@@ -413,8 +451,8 @@ BayesianOptimization <- function(
   }
 
   RetList <- list()
-  names(GPlist) <- paste0("GP_", 1:Iter)
-  RetList$GPlist <- GPlist
+  RetList$GPs <- GPs
+  if (!is.null(GPe)) RetList$GPe <- GPe
   RetList$acqMaximums <- acqMaximums
   RetList$ScoreDT <- ScoreDT
   RetList$BestPars <- BestPars
